# Module with most used assertions and checks
import numpy as np
from math import isnan


def is_na(var: object) -> bool:
    na = var is None or var is np.nan
    try:
        na = na or isnan(var)
    except TypeError:
        pass
    return na


if __name__ == "__main__":
    print(is_na(None))
    print(is_na(np.nan))
    print(is_na(float('nan')))
    print(is_na('x'))
    print(is_na(''))



import time
from contextlib import contextmanager
from datetime import datetime
from typing import Iterator

# Simplest - generator-iterator based
@contextmanager
def timer(code_block_name: str) -> Iterator[None]:
    """Prints the time taken to execute a given code block.

    Args:
        code_block_name (str): Description of your codeblock

    Usage:
        with timer("for loop 100,000,000"):
            _ = [x for x in range(100_000_000)]

        prints #=> [for loop 100,000,000] completed in 4 seconds.. 
    """
    start_time = time.time()
    # yields the control to the with statement
    yield
    # executed after with code block execution
    execution_time = time.time() - start_time
    print(f"[{code_block_name}] completed in {execution_time:.0f} seconds..")


# Return Start Date and time
# just to show the use of `as` assignment in the `with` statement
@contextmanager
def timer_v2(code_block_name: str) -> Iterator[datetime]:
    """Prints the time taken to execute a given code block.

    Args:
        code_block_name (str): Description of your codeblock

    Yields:
        str: Execution start date time ISO 8601 format
    """
    start_time = time.time()
    # yields the value to the with statement
    yield datetime.now()
    # executed after with code block execution
    execution_time = time.time() - start_time
    print(f"[{code_block_name}] completed in {execution_time:.0f} seconds..")

if __name__ == "__main__":
    # sample use
    with timer("for loop 100,000,000"):
        _ = [x for x in range(100_000_000)]


# major.minor.patch
__version__ = "0.1"

from .utils.timing import timer
from .utils.parallel import mp_func
from .utils.save_load import (
    save_json,
    load_json,
    load_pickle,
    save_pickle,
    assert_file_exists,
    assert_dir_exists,
)
from .data.pd_ops import (
    assert_column_in_dataframe,
    df_get_dummy,
    df_filter_values_above_freq,
    series_most_freq,
)




import logging
from os import listdir
from os.path import isfile, join, isdir, dirname
import os
import shutil
from tqdm import tqdm
from typing import Dict, List


class ManageDir:
    """Commonly used directory functions to manage a directory."""

    suffix = "_"

    @staticmethod
    def assert_file_exists(file_path: str) -> None:
        """Assert filepath exists. Give verbose error messages.

        Args:
            file_path (str): file path e.g. abc/xyz.csv or xyz.csv
        """
        dir = dirname(file_path)
        assert dir == "" or isdir(dir), f"{dir} directory doesn't exist.."
        assert isfile(
            file_path
        ), f"{file_path} is invalid. \nContents of {dir} are {[f for f in listdir(dir)]}"

    @staticmethod
    def assert_parent_dir_exists(file_path: str) -> None:
        """Assert parent directory of a file_path exists

        Args:
            file_path (str): file path e.g. abc/xyz.csv or xyz.csv
        """
        dir = dirname(file_path)
        assert dir == "" or isdir(dir), f"{dir} directory doesn't exist.."

    @staticmethod
    def assert_dir_exists(dir_path: str) -> None:
        """Assert directory exists

        Args:
            dir_path (str): dir path e.g. abc/ or abc/xyz/
        """
        dir = dirname(dir_path)
        assert dir == "" or isdir(dir), f"{dir} directory doesn't exist.."
        assert isdir(dir_path), f"{dir_path} directory doesn't exist.."

    @classmethod
    def list_empty_sub_dirs(cls, dir_path: str) -> List[str]:
        """Get empty dirs in given directory path.

        Args:
            dir_path (str): String directory path

        Returns:
            List[str]: List of sub directory path if those dirs are empty
        """

        cls.assert_dir_exists(dir_path)
        return [
            join(dir_path, f)
            for f in listdir(dir_path)
            if isdir(join(dir_path, f)) and not listdir(join(dir_path, f))
        ]

    @classmethod
    def remove_empty_sub_dirs(cls, dir_path: str, verbose: bool = False) -> None:
        """Removes empty directories from provided dir path.

        Args:
            dir_path (str): String directory path
            verbose (bool, optional): print out name of deleted dir. Defaults to False.
        """
        for dir in cls.list_empty_sub_dirs(dir_path):
            os.rmdir(dir)
            if verbose:
                logging.info(f"Deleted {dir}")

    @staticmethod
    def get_filetype(file_path: str) -> str:
        """Get fie type from extension.

        Args:
            file_path (str): filepath

        Returns:
            str: filetype, eg. for `abc.csv` return `csv`
        
        Examples:
            xy.jpg      -> jpg
            .gitignore  -> unk
        """
        _, filetype = os.path.splitext(file_path)

        if not filetype:
            filetype = "unk"

        filetype = filetype.replace(".", "")
        return filetype

    @classmethod
    def get_type_sub_dirs(cls, dir_path: str) -> Dict[str, str]:
        """Get sub dirs which are created by us

        Args:
            dir_path (str): dir to search into

        Returns: 
            Dict[str, str]: filetype to file type sub directory path
        """
        cls.assert_dir_exists(dir_path)

        return {
            f[:-1]: join(dir_path, f)
            for f in listdir(dir_path)
            if isdir(join(dir_path, f)) and str(f).endswith(cls.suffix)
        }

    @classmethod
    def create_type_sub_dirs(cls, dir_path: str) -> Dict[str, str]:
        """Creates type sub folders (if doesn't exist).

        Args:
            dir_path (str): directory path to create into

        Returns:
            Dict[str, str]: filetype to file type sub dirs path
        """

        # check if the directory to arrange exists
        cls.assert_dir_exists(dir_path)

        dir_path = os.path.abspath(dir_path)

        logging.info("Working on {}\n".format(dir_path))

        files = [f for f in listdir(dir_path) if isfile(join(dir_path, f))]

        # previously created file_type folders
        filetype_folder_dict = cls.get_type_sub_dirs(dir_path)

        # creating directory
        for file in tqdm(files):
            filetype = cls.get_filetype(file)

            if filetype not in filetype_folder_dict:
                new_file_type_folder = join(
                    dir_path, "{}{}".format(filetype, cls.suffix)
                )

                # if folder doesn't exists -> create folder for file type
                if not isdir(new_file_type_folder):
                    os.mkdir(new_file_type_folder)
                    filetype_folder_dict[str(filetype)] = str(new_file_type_folder)

        return filetype_folder_dict

    @classmethod
    def move_files_to_type_sub_dirs(cls, dir_path: str, verbose: bool = False) -> None:
        """Creates type folders in a directory and moves files to respective type folders.

        Args:
            dir_path (str): Directory to work upon
            verbose (bool, optional): If true log INFO. Defaults to False.
        """
        cls.assert_dir_exists(dir_path)
        filetype_folder_dict = cls.create_type_sub_dirs(dir_path)
        dir_path = os.path.abspath(dir_path)
        logging.info("Working on {}\n".format(dir_path))
        files = [f for f in listdir(dir_path) if isfile(join(dir_path, f))]
        # moving files
        for file in tqdm(files):
            src_path = os.path.join(dir_path, file)
            filetype = cls.get_filetype(file)
            if filetype in filetype_folder_dict.keys():
                dest_folder = filetype_folder_dict[str(filetype)]
                dest_path = os.path.join(dest_folder, file)
                shutil.move(src_path, dest_path)
                if verbose:
                    logging.info(f"{src_path} --to---> {dest_path}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="OCD args")
    parser.add_argument(
        "-D", "--dir", type=str, required=True, nargs=1, help="directory to run on",
    )

    # example args
    # args = parser.parse_args("-D /Users/userx/Downloads".split())

    # args from cli
    args = parser.parse_args()

    path = args.dir[0]
    # print(f"Empty in {path}: {ManageDir.list_empty_sub_dirs(path)}")
    # ManageDir.remove_empty_sub_dirs(path)
    ManageDir.move_files_to_type_sub_dirs(path)



import pickle
import json
from os.path import isdir, isfile, dirname
from posix import listdir
from typing import Dict, Hashable


def load_pickle(pickle_path: str) -> object:
    """Avoid boilerplate pickle loading.

    Args:
        pickle_path (str): Path of pickle file

    Returns:
        object: Python object unpickled
    """
    assert_file_exists(pickle_path)

    with open(pickle_path, "rb") as pf:
        return pickle.load(pf)


def save_pickle(obj: object, pickle_path: str) -> None:
    """Avoid boilerplate pickle saving.

    Args:
        obj (object): desired python object
        pickle_path (str): Path of pickle file
    """
    assert_dir_exists(pickle_path)
    with open(pickle_path, "wb") as pf:
        # protocol 4 is default in python 3.8
        pickle.dump(obj, pf, protocol=4)


def load_json(json_path: str, key_is_int: bool = False) -> Dict[Hashable, object]:
    """Avoid boilerplate json loading.

    Args:
        json_path (str): path to json file
        key_is_int (bool, optional): Convert keys to int. Defaults to False.

    Returns:
        object: a python dict object
    """

    assert_file_exists(json_path)

    with open(json_path, "r") as jf:
        if key_is_int:
            return {int(key): value for key, value in json.load(jf).items()}
        return json.load(jf)


def save_json(obj: Dict[Hashable, object], json_path: str) -> None:
    """Save python dict to json_path

    Args:
        obj (Dict[Hashable, object]): python dict
        json_path (str): path
    """

    assert_dir_exists(json_path)

    with open(json_path, "w") as jf:
        json.dump(obj, jf)


def assert_file_exists(file_path: str) -> None:
    """Assert filepath exists. Give verbose error messages.

    Args:
        file_path (str): file path e.g. abc/xyz.csv or xyz.csv
    """
    dir = dirname(file_path)
    assert dir == "" or isdir(dir), f"{dirname} directory doesn't exist.."
    assert isfile(
        file_path
    ), f"{file_path} is invalid. \nContents of {dir} are {[f for f in listdir(dir)]}"


def assert_dir_exists(file_path: str) -> None:
    """Assert directory of a file_path exists

    Args:
        file_path (str): file path e.g. abc/xyz.csv or xyz.csv
    """
    dir = dirname(file_path)
    assert dir == "" or isdir(dir), f"{dirname} directory doesn't exist.."




from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")


setup(
    name="boondh",
    version="0.1.2",
    description="A Swiss Knife",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["utility", "data", "multiprocessing", "cli", "tools"],
    url="https://github.com/rohit-mehra/boondh",
    author="Rohit Mehra",
    author_email="iirohitmehra@gmail.com",
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=[
        "numpy",
        # dataclasses for Python versions that don't have it
        'dataclasses;python_version<"3.7"',
        # utilities from PyPA to e.g. compare versions
        "packaging",
        # progress bars
        "tqdm >= 4.27",
        # data ops
        "pandas",
    ],
    python_requires=">=3.6.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
    ],
)




from traceback import format_exc
from tqdm import tqdm
from math import sqrt
from inspect import getfullargspec
from multiprocessing import Pool, cpu_count
from functools import partial
from typing import Iterable, Union
import logging
log = logging.getLogger(__name__)

def mp_func(
    func,
    data_arg_name_in_func: str,
    data: Iterable,
    chunksize: Union[str, int] = "auto",
    *args,
    **kwargs,
):
    """Apply given function on each element of data in parallel and show progress bar..

    Args:
        func: the python function to be applied
        data_arg_name_in_func (str): argument name in `func` which requires data point from data iterable
        data (Iterable): iterable with data points
        chunksize (str, int): 'auto' or None for automatic calc, int for specific chunksize
        args: positional args which will be supplied to the given `func`
        kwargs: keyword args which will be supplied to the given `func`


    Returns:
        list: result with each value as a result of application of function `func` 
        on each point in the given `data` iterable

                                                               
    Example:
    To parallelize given `base_func` function: 
        >>> def base_func(value, sq=True):
                if sq: return value ** 2
                return value
        >>> data = [0, 1, 2, 3, 4]
        >>> results = mp_func(base_func, 'value', data, sq=True)
        >>> print(results)
        [0, 1, 4, 9, 16]
        >>> results = mp_func(base_func, 'value', data, sq=False)
        >>> print(results)
        [0, 1, 2, 3, 4]
    """

    func_args = getfullargspec(func).args

    # Sanity checks
    assert isinstance(data, Iterable), "data should be an iterable.."
    total = len(data)
    assert (
        data_arg_name_in_func in func_args
    ), f"{data_arg_name_in_func} is not an argument of {func.__name__} function that you provided.."
    assert total > 1, f"len(data) should be > 1, but is {len(data)}"
    assert len(args) + len(kwargs) + 1 == len(
        func_args
    ), f"{len(args)} + {len(kwargs)} + 1 != {len(func_args)}\nCheck args func_args are {func_args}"

    # for better user experience
    cpus = cpu_count()
    log.info(f"CPUS: {cpus}")

    # Returns a new partial object which when called will behave
    # like func called with the positional arguments args and keyword arguments kwargs.
    _new_func = partial(func, *args, **kwargs)
    results = None

    try:
        # sanity check
        _ = _new_func(**{data_arg_name_in_func: data[-1]})

        if not chunksize or isinstance(chunksize, str):
            if chunksize == "auto":
                chunksize = int(sqrt(total) * cpus / 2)
            else:
                logging.warning(f"chunksize can't be {chunksize} use 'auto' or any int value..using auto..")
                chunksize = int(sqrt(total) * cpus / 2)
        log.info(f"chunksize: {chunksize}")
                
        # actual processing
        with Pool(cpus) as pool:
            results = list(
                tqdm(pool.imap(_new_func, data, chunksize=chunksize), total=total)
            )
    except TypeError:
        log.info(
            f"Please make sure that args and kwargs provided are valid for {func.__name__} function.."
        )
        log.exception(format_exc())
    except Exception:
        log.exception(format_exc())
    finally:
        log.info("Done..")

    return results


if __name__ == "__main__":
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    log.info("Starting mp_func on 100_000_000 sample points..")
    # Sample Usage
    # function to be applied
    def base_func(value, sq=True):
        if sq:
            return value ** 2
        return value

    results = mp_func(base_func, "value", data=range(0, 100_000_000), chunksize=200, sq=True)
    print(results[33])



from pandas import DataFrame, Series
from numpy import nan
from string import ascii_uppercase
from random import randint


def series_most_freq(series: Series) -> object:
    """Returns most feq value in the series. Good as agg function.

    Args:
        series (Series): Pandas series

    Returns:
        object: Either the series value or numpy.nan
    """
    series = series.dropna().values.tolist()
    if series:
        return max(series, key=series.count)
    return nan


def df_filter_values_above_freq(
    df: DataFrame, column_name: str, freq: int
) -> DataFrame:
    """Filters data frame on a column's value counts.

    Args:
        df (DataFrame): DataFrame to be filtered
        column_name (str): target column
        freq (int): minimum freq (inclusive) of values in that column to be accepted in new dataframe

    Returns:
        DataFrame: DataFrame with only those rows which have values that occur in that column more than `freq` times
    """
    assert_column_in_dataframe(column_name, df)
    value_counts = df[column_name].value_counts()
    values_with_freq_above = value_counts.loc[value_counts >= freq].index.values
    return df.loc[df[column_name].isin(values_with_freq_above)]


def df_get_dummy() -> DataFrame:
    """Create a dummy dataframe for experiments..

    Returns:
        DataFrame: DataFrame of shape (26, 3)
    """
    sales = [randint(0, 100) for _ in range(len(ascii_uppercase))]
    category = [sale // 10 for sale in sales]
    data = DataFrame(
        {"name": list(ascii_uppercase), "sales": sales, "category": category,}
    )
    return data.sample(data.shape[0]).reset_index(drop=True)


def assert_column_in_dataframe(column_name: str, df: DataFrame):
    """Assert column in a dataframe

    Args:
        column_name (str): desired column
        df (DataFrame): target dataframe
    """
    assert (
        column_name in df.columns.values
    ), f"{column_name} not in DataFrame\nColumns={df.columns.values.tolist()}"


if __name__ == "__main__":
    df = df_get_dummy()
    print(df_filter_values_above_freq(df, "category", 5))
    print(series_most_freq(Series([1,2,3,4,1,2,3,2,3,2])))
